#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef rti_me_cpp_hxx
#include "rti_me_cpp.hxx"
#endif
#include "wh_sm/wh_sm_history.h"
#include "rh_sm/rh_sm_history.h"

#include "HelloWorld.h"
#include "HelloWorldSupport.h"
#include "HelloWorldPlugin.h"
#include "HelloWorldApplication.h"


#define PEER_NAME       "subscriber"
struct ApplicationSecMaterial PEER_MATERIAL_DEFAULT;

class HelloWorldReaderListener : public DDSDataReaderListener
{

  public:

    virtual void on_data_available(DDSDataReader *reader);

    virtual DDS_Boolean on_before_sample_deserialize(
            DDSDataReader *reader,
            NDDS_Type_Plugin *plugin,
            CDR_Stream_t *stream,
            DDS_Boolean *dropped);

    virtual DDS_Boolean on_before_sample_commit(
            DDSDataReader *reader,
            const void *const sample,
            const struct DDS_SampleInfo *const sample_info,
            DDS_Boolean *dropped);

  private:
    virtual void filter_sample(HelloWorld *sample,
                               DDS_Boolean *drop_sample);

    virtual void deserialize_unsigned_long(char **src_buffer,
                                           RTI_BOOL need_byte_swap,
                                           DDS_UnsignedLong *instance);
};


template <typename T>
void take_and_print(typename T::DataReader* reader)
{
    DDS_ReturnCode_t retcode;
    struct DDS_SampleInfo sample_info;
    T *sample = NULL;

    sample = T::TypeSupport::create_data();
    if (sample == NULL)
    {
        printf("Failed sample initialize\n");
        return;
    }

    retcode = reader->take_next_sample(*sample, sample_info);
    while (retcode == DDS_RETCODE_OK)
    {
        if (sample_info.valid_data)
        {
            printf("\nSample received\n\tmsg: %s\n", sample->msg);
        }
        else
        {
            printf("\nSample received\n\tINVALID DATA\n");
        }
        retcode = reader->take_next_sample(*sample, sample_info);
    }

    T::TypeSupport::delete_data(sample);
}

void
HelloWorldReaderListener::on_data_available(DDSDataReader * reader)
{
    HelloWorldDataReader *hw_reader = HelloWorldDataReader::narrow(reader);
    take_and_print<HelloWorld>(hw_reader);
}

void
HelloWorldReaderListener::filter_sample(HelloWorld *sample,
                                        DDS_Boolean *drop_sample)
{
    /* Example filter: drop samples with even-numbered count in id */

    *drop_sample = (sample->id % 2 == 0) ?
                        DDS_BOOLEAN_TRUE : DDS_BOOLEAN_FALSE;
}

void
HelloWorldReaderListener::deserialize_unsigned_long(char **src_buffer,
                                                    RTI_BOOL need_byte_swap,
                                                    DDS_UnsignedLong *instance)
{
    RTI_INT32 i;
    if (need_byte_swap)
    {
        for (i = 3; i >= 0; --i)
        {
            *((RTI_INT8*)instance + i) = *((*src_buffer)++);
        }
    }
    else
    {
        *instance = *(RTI_UINT32*)(*src_buffer);
        (*src_buffer) += CDR_LONG_SIZE;
    }
}

DDS_Boolean
HelloWorldReaderListener::on_before_sample_deserialize(
            DDSDataReader *reader,
            NDDS_Type_Plugin *plugin,
            CDR_Stream_t *stream,
            DDS_Boolean *dropped)
{
/* Unsigned Long type size */
#define UNSIGNED_LONG_SIZE  4

#ifdef USE_SAMPLE_FILTER
#ifdef FILTER_ON_DESERIALIZE
    DDS_Boolean result = DDS_BOOLEAN_FALSE;
    DDS_Long id = 0;
    RTI_BOOL need_byte_swap = DDS_BOOLEAN_FALSE;
    char *src_buffer = NULL;

    need_byte_swap = CDR_Stream_is_byte_swapped(stream);
    src_buffer = CDR_Stream_get_current_position_ptr(stream);

    /* deserialize only field 'id', instead of the whole sample */

    if (!CDR_Stream_check_size(stream, UNSIGNED_LONG_SIZE))
    {
        printf("Failed to deserialize id. The stream is too short, missing data\n");
        goto done;
    }

    /* Note primitive types must be aligned to their length in the CDR stream.
     * For example, a long must start on a 4-byte boundary. The boundaries are
     * counted from the start of the CDR stream.
     * As the sample 'id' is the first data in the stream it is already aligned.
     * Position 0 (beginning of the stream) is aligned to 4 (size of long).
     */
    this->deserialize_unsigned_long(&src_buffer,
                                    need_byte_swap,
                                    (DDS_UnsignedLong*)&id);

    /* filter the sample */
    *dropped = (id % 2 == 0) ? DDS_BOOLEAN_TRUE : DDS_BOOLEAN_FALSE;

    if (*dropped)
    {
        printf("\nSample filtered, before deserialize...\n\tDROPPED - id: %d\n",
               id);
    }

    result = DDS_BOOLEAN_TRUE;
done:
    return result;
#endif
#else
    *dropped = RTI_FALSE;
    return DDS_BOOLEAN_TRUE;
#endif
}

DDS_Boolean
HelloWorldReaderListener::on_before_sample_commit(
            DDSDataReader *reader,
            const void *const sample,
            const struct DDS_SampleInfo *const sample_info,
            DDS_Boolean *dropped)
{
#ifdef USE_SAMPLE_FILTER
#ifndef FILTER_ON_DESERIALIZE
    HelloWorld *hw_sample = (HelloWorld *)sample;

    this->filter_sample(hw_sample, dropped);

    if (*dropped)
    {
        printf("\nSample filtered, before commit...\n\tDROPPED - msg: %s\n",
               hw_sample->msg);
    }
#endif
#endif
    return DDS_BOOLEAN_TRUE;
}

int
subscriber_main_w_args(DDS_Long domain_id, char *udp_intf, char *peer,
                       DDS_Long sleep_time, DDS_Long count,
                       ApplicationSecMaterial *sec_material)
{
    DDSSubscriber *subscriber = NULL;
    DDSDataReader *datareader = NULL;
    HelloWorldReaderListener *listener  = new HelloWorldReaderListener();
    DDS_DataReaderQos dr_qos;
    DDS_ReturnCode_t retcode;
    Application *application = NULL;
    RT_Registry_T *registry = NULL;

    application = new Application();
    if (application == NULL)
    {
        printf("failed Application new\n");
        goto done;
    }

    retcode = application->initialize(PEER_NAME,
                                      "publisher", 
                                      domain_id,
                                      udp_intf, 
                                      peer, 
                                      sleep_time, 
                                      count,
                                      sec_material);
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed Application initialize\n");
        goto done;
    }

    subscriber = application->participant->create_subscriber(
                        DDS_SUBSCRIBER_QOS_DEFAULT,NULL,
                        DDS_STATUS_MASK_NONE);
    if (subscriber == NULL)
    {
        printf("subscriber == NULL\n");
        goto done;
    }

    retcode = subscriber->get_default_datareader_qos(dr_qos);
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed get_default_datareader_qos\n");
        goto done;
    }

/* Publisher sends samples with id = 0 or id = 1, so 2 instances maximum.
     * But in case filtering is done, all samples with 'id = 0' are
     * filtered so only one instance is needed.
     */
#ifdef USE_SAMPLE_FILTER
    dr_qos.resource_limits.max_instances = 1;
#else
    dr_qos.resource_limits.max_instances = 2;
#endif

    dr_qos.resource_limits.max_samples_per_instance = 32;
    dr_qos.resource_limits.max_samples =
            dr_qos.resource_limits.max_instances *
                            dr_qos.resource_limits.max_samples_per_instance;
    /* if there are more remote writers, you need to increase these limits */
    dr_qos.reader_resource_limits.max_remote_writers = 10;
    dr_qos.reader_resource_limits.max_remote_writers_per_instance = 10;
    dr_qos.history.depth = 32;

    /* Reliability QoS */
#ifdef USE_RELIABLE_QOS
    dr_qos.reliability.kind = DDS_RELIABLE_RELIABILITY_QOS;
#else
    dr_qos.reliability.kind = DDS_BEST_EFFORT_RELIABILITY_QOS;
#endif

    /* Only DDS_DATA_AVAILABLE_STATUS supported currently */
    datareader = subscriber->create_datareader(
                           application->topic,
                           dr_qos,
                           listener,
                           DDS_DATA_AVAILABLE_STATUS);

    if (datareader == NULL)
    {
        printf("datareader == NULL\n");
        goto done;
    }

    retcode = application->enable();
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed to enable application\n");
        goto done;
    }

    if (application->count != 0)
    {
        printf("Running for %d seconds, press Ctrl-C to exit\n",
               application->count);
        OSAPI_Thread_sleep(application->count * 1000);
    }
    else
    {
        int sleep_loop_count =  (24 * 60 * 60) / 2000;
        int sleep_loop_left = (24 * 60 * 60) % 2000;

        printf("Running for 24 hours, press Ctrl-C to exit\n");

        while (sleep_loop_count)
        {
            OSAPI_Thread_sleep(2000  * 1000); /* sleep is in ms */
            --sleep_loop_count;
        }

        OSAPI_Thread_sleep(sleep_loop_left * 1000);
    }

    done:

    if (application != NULL)
    {
        delete application;
    }
   
    if (listener != NULL)
    {
        delete listener;
    }

    //DataReaderQos is automatically disposed by destructor

    return 0;
}

#if !(defined(RTI_VXWORKS) && !defined(__RTP__))
int
main(int argc, char **argv)
{
    DDS_Long i = 0;
    DDS_Long domain_id = 0;
    char *peer = NULL;
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;
    struct ApplicationSecMaterial sec_material(
        /* ca_id_cert */   ca_string,
        /* ca_perm_cert */ ca_string,
        /* peer_key */     subscriber_key_string,
        /* peer_cert */    subscriber_string,
        /* xml_gov */      governance_string,
        /* xml_perm */     permissions_subscriber_string);
    PEER_MATERIAL_DEFAULT = sec_material;

    for (i = 1; i < argc; ++i)
    {
        if (!strcmp(argv[i], "-domain"))
        {
            ++i;
            if (i == argc)
            {
                printf("-domain <domain_id>\n");
                return -1;
            }
            domain_id = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-udp_intf"))
        {
            ++i;
            if (i == argc)
            {
                printf("-udp_intf <interface>\n");
                return -1;
            }
            udp_intf = argv[i];
        }
        else if (!strcmp(argv[i], "-peer"))
        {
            ++i;
            if (i == argc)
            {
                printf("-peer <address>\n");
                return -1;
            }
            peer = argv[i];
        }
        else if (!strcmp(argv[i], "-sleep"))
        {
            ++i;
            if (i == argc)
            {
                printf("-sleep_time <sleep_time>\n");
                return -1;
            }
            sleep_time = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-count"))
        {
            ++i;
            if (i == argc)
            {
                printf("-count <count>\n");
                return -1;
            }
            count = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-ca-id"))
        {
            ++i;
            if (i == argc)
            {
                printf("-ca-id <ca.pem>\n");
                return -1;
            }
            sec_material.ca_id_cert = argv[i];
        }
        else if (!strcmp(argv[i], "-ca-perm"))
        {
            ++i;
            if (i == argc)
            {
                printf("-ca-perm <ca.pem>\n");
                return -1;
            }
            sec_material.ca_perm_cert = argv[i];
        }
        else if (!strcmp(argv[i], "-key"))
        {
            ++i;
            if (i == argc)
            {
                printf("-key <peer_key.pem>\n");
                return -1;
            }
            sec_material.peer_key = argv[i];
        }
        else if (!strcmp(argv[i], "-cert"))
        {
            ++i;
            if (i == argc)
            {
                printf("-cert <peer.pem>\n");
                return -1;
            }
            sec_material.peer_cert = argv[i];
        }
        else if (!strcmp(argv[i], "-gov"))
        {
            ++i;
            if (i == argc)
            {
                printf("-gov <governance.p7s>\n");
                return -1;
            }
            sec_material.xml_gov = argv[i];
        }
        else if (!strcmp(argv[i], "-perm"))
        {
            ++i;
            if (i == argc)
            {
                printf("-perm <permissions.p7s>\n");
                return -1;
            }
            sec_material.xml_perm = argv[i];
        }
        else if (!strcmp(argv[i], "-h"))
        {
            Application::help(argv[0]);
            return 0;
        }
        else
        {
            printf("unknown option: %s\n", argv[i]);
            return -1;
        }
    }

    return subscriber_main_w_args(
                domain_id, udp_intf, peer, sleep_time, count, &sec_material);
}
#elif defined(RTI_VXWORKS)
int
subscriber_main(void)
{
    /* Explicitly configure args below */
    DDS_Long domain_id = 44;
    char *peer = "10.10.65.103";
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;
    struct ApplicationSecMaterial sec_material = PEER_MATERIAL_DEFAULT;

    return subscriber_main_w_args(
                domain_id, udp_intf, peer, sleep_time, count, &sec_material);
}
#endif


